from blib import Query
from datetime import datetime, timedelta

from .base import ClientBase, object_to_id, object_to_ids, parse_data

from ..base import ApiDate
from ..enums import StatusAction, VisibilityLevel
from ..objects import (
	Account,
	ListDeserializer,
	Media,
	ScheduledStatus,
	Status,
	StatusContext,
	StatusEdit,
	StatusSource,
	Translation
)


class StatusBase(ClientBase):
	async def account_statuses(self,
								acct: Account | int | str,
								min_id: int | None = None,
								max_id: int | None = None,
								since_id: int | None = None,
								limit: int = 20, # max 40
								only_media: bool = False,
								exclude_replies: bool = False,
								exclude_reblogs: bool = False,
								pinned: bool = False,
								tag: str | None = None) -> list[Status]:

		acct = object_to_id(acct)
		data = parse_data(
			limit = limit,
			only_media = only_media,
			exclude_replies = exclude_replies,
			exclude_reblogs = exclude_reblogs,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			tagged = tag
		)

		return await self.send(
			"GET", f"/api/v1/accounts/{acct}/statuses", ListDeserializer(Status), data
		)


	async def bookmark(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.BOOKMARK)


	async def bookmarks(self,
					min_id: int | None = None,
					max_id: int | None = None,
					since_id: int | None = None,
					limit: int = 20) -> list[Status]:

		query = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send("GET", "/api/v1/bookmarks", ListDeserializer(Status), query = query)


	async def boost(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.BOOST)


	async def boosted_by(self, status: Status | int | str) -> list[Account]:
		status = object_to_id(status)
		return await self.send(
			"GET", f"/api/v1/statuses/{status}/reblogged_by", ListDeserializer(Account)
		)


	async def cancel_scheduled_status(self, status: Status | int | str) -> None:
		status = object_to_id(status)
		return await self.send("DELETE", f"/api/v1/scheduled_statuses/{status}", None)


	async def delete_status(self, status: Status | int | str) -> None:
		status = object_to_id(status)
		return await self.send("DELETE", f"/api/v1/statuses/{status}", None)


	async def favorite(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.FAVORITE)


	async def favorited_by(self, status: Status | int | str) -> list[Account]:
		status = object_to_id(status)
		return await self.send(
			"GET", f"/api/v1/statuses/{status}/favourited_by", ListDeserializer(Account)
		)


	async def favorites(self,
						min_id: int | None = None,
						max_id: int | None = None,
						since_id: int | None = None,
						limit: int = 20) -> list[Status]:

		query = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send("GET", "/api/v1/favourites", ListDeserializer(Status), query = query)


	async def mute(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.MUTE)


	async def new_scheduled_status(self,
					text: str,
					summary: str | None = None,
					reply_to: Status | int | str | None = None,
					media: list[Media | int | str] | None = None,
					sensitive: bool = False,
					poll_options: list[str] | None = None,
					poll_expires: timedelta | int | None = None,
					poll_multiple: bool = False,
					poll_hide_totals: bool = False,
					level: VisibilityLevel | None = None,
					language: str | None = None,
					scheduled_at: datetime | None = None,
					idempotency_key: str | None = None) -> Status:
		return Status()


	async def new_status(self,
					text: str,
					summary: str | None = None,
					reply_to: Status | int | str | None = None,
					media: list[Media | int | str] | None = None,
					sensitive: bool = False,
					poll_options: list[str] | None = None,
					poll_expires: timedelta | int | None = None,
					poll_multiple: bool = False,
					poll_hide_totals: bool = False,
					level: VisibilityLevel | None = None,
					language: str | None = None,
					scheduled_at: datetime | None = None,
					idempotency_key: str | None = None) -> Status:

		poll = None

		if poll_options:
			if isinstance(poll_expires, timedelta):
				poll_expires = poll_expires.seconds

			poll = parse_data(
				options = poll_options,
				expires_in = poll_expires,
				multiple = poll_multiple,
				hide_totals = poll_hide_totals
			)

		if reply_to is not None:
			reply_to = object_to_ids([reply_to])[0]

		data = parse_data(
			status = text,
			media_ids = object_to_ids(media or []) or None,
			poll = poll,
			in_reply_to_id = reply_to,
			sensitive = sensitive,
			spoiler_text = summary,
			language = language,
			visibility = VisibilityLevel.parse(level) if level is not None else level,
			scheduled_at = scheduled_at.isoformat() if scheduled_at is not None else scheduled_at
		)

		return await self.send("POST", "/api/v1/statuses", Status, data)


	async def pin(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.PIN)


	async def scheduled_status(self, id: int | str) -> ScheduledStatus:
		return await self.send("GET", f"/api/v1/scheduled_statuses/{id}", ScheduledStatus)


	async def scheduled_statuses(self,
								min_id: int | None = None,
								max_id: int | None = None,
								since_id: int | None = None,
								limit: int = 20) -> list[ScheduledStatus]:

		data = parse_data(
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			limit = limit
		)

		return await self.send(
			"GET", "/api/v1/scheduled_statuses", ListDeserializer(ScheduledStatus), data
		)


	async def status(self, status_id: int | str) -> Status:
		return await self.send("GET", f"/api/v1/statuses/{status_id}", Status)


	async def status_action(self, status: Status | int | str, action: StatusAction | str) -> Status:
		status = object_to_id(status)
		action = StatusAction.parse(action)
		return await self.send("POST", f"/api/v1/statuses/{status}/{action}", Status)


	async def status_context(self, status: Status | int | str) -> StatusContext:
		status = object_to_id(status)
		return await self.send("GET", f"/api/v1/statuses/{status}/context", StatusContext)


	async def status_history(self, status: Status | int | str) -> list[StatusEdit]:
		status = object_to_id(status)
		return await self.send(
			"GET", f"/api/v1/statuses/{status}/history", ListDeserializer(StatusEdit)
		)


	async def status_source(self, status: Status | int | str) -> StatusSource:
		status = object_to_id(status)
		return await self.send("GET", f"/api/v1/statuses/{status}/source", StatusSource)


	async def statuses(self, *ids: int | str) -> list[Status]:
		data = Query()

		for item in ids:
			data.append("id[]", str(item))

		return await self.send("GET", "/api/v1/statuses", ListDeserializer(Status), query = data)


	async def translate_status(self,
							status: Status | int | str,
							lang: str | None = None) -> Translation:

		status = object_to_id(status)
		data = parse_data(lang = lang)
		return await self.send("POST", f"/api/v1/statuses/{status}/translate", Translation, data)


	async def unbookmark(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.UNBOOKMARK)


	async def unboost(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.UNBOOST)


	async def unfavorite(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.UNFAVORITE)


	async def unmute(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.UNMUTE)


	async def unpin(self, status: Status | int | str) -> Status:
		return await self.status_action(status, StatusAction.UNPIN)


	async def update_scheduled_status_date(self,
										status: Status | int | str,
										date: ApiDate | str) -> ScheduledStatus:

		status = object_to_id(status)
		data = {"scheduled_at": ApiDate.parse(date).to_string()}
		return await self.send("PUT", f"/api/v1/scheduled_statuses/{status}", ScheduledStatus, data)


	async def update_status(self,
					status: Status | int | str,
					text: str | None = None,
					summary: str | None = None,
					reply_to: Status | int | str | None = None,
					media: list[Media | int | str] | None = None,
					sensitive: bool = False,
					poll_options: list[str] | None = None,
					poll_expires: timedelta | int | None = None,
					poll_multiple: bool = False,
					poll_hide_totals: bool = False,
					level: VisibilityLevel | None = None,
					language: str | None = None,
					scheduled_at: datetime | None = None,
					idempotency_key: str | None = None) -> Status:

		status = object_to_id(status)
		poll = None

		if poll_options:
			if isinstance(poll_expires, timedelta):
				poll_expires = poll_expires.seconds

			poll = parse_data(
				options = poll_options,
				expires_in = poll_expires,
				multiple = poll_multiple,
				hide_totals = poll_hide_totals
			)

		if reply_to is not None:
			reply_to = object_to_ids([reply_to])[0]

		data = parse_data(
			status = text,
			media_ids = object_to_ids(media or []) or None,
			poll = poll,
			in_reply_to_id = reply_to,
			sensitive = sensitive,
			spoiler_text = summary,
			language = language,
			visibility = VisibilityLevel.parse(level) if level is not None else level,
			scheduled_at = scheduled_at.isoformat() if scheduled_at is not None else scheduled_at
		)

		return await self.send("PUT", f"/api/v1/statuses/{status}", Status, data)


	async def merp(self, text: str, summary: str | None = None) -> Status:
		"""
			Quickly blast your nonsense to the fediverse

			:param text: A random thought to bother the fediverse with
		"""

		data = parse_data(
			status = text,
			spoiler_text = summary
		)

		return await self.send("POST", "/api/v1/statuses", Status, data)



from collections.abc import Sequence

from .base import ClientBase, object_to_id, parse_data

from ..enums import NotificationType
from ..objects import (
	Account,
	ListDeserializer,
	Notification,
	NotificationPolicy,
	NotificationRequest
)


class NotificationBase(ClientBase):
	# missing api endpoint in docs, so this is just an educated guess
	async def accept_notification_request(self, notif: Notification | int | str) -> None:
		notif = object_to_id(notif)
		return await self.send("POST", f"/api/v1/notifications/requests/:{notif}/accept", None)


	async def deny_notification_request(self, notif: Notification | int | str) -> None:
		notif = object_to_id(notif)
		return await self.send("POST", f"/api/v1/notifications/requests/:{notif}/dismiss", None)


	async def notification(self, id: int | str) -> Notification:
		return await self.send("GET", f"/api/v1/notifications/{id}", Notification)


	async def notification_policy(self) -> NotificationPolicy:
		return await self.send("GET", "/api/v1/notification/policy", NotificationPolicy)


	async def notification_request(self, id: int | str) -> NotificationRequest:
		return await self.send("GET", f"/api/v1/notifications/requests/{id}", NotificationRequest)


	async def notification_requests(self,
									min_id: int | None = None,
									max_id: int | None = None,
									since_id: int | None = None,
									limit: int = 40,
									dismissed: bool = False) -> list[NotificationRequest]:

		data = parse_data(
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			limit = limit,
			dismissed = dismissed
		)

		return await self.send(
			"GET", "/api/v1/notifications/requests", ListDeserializer(NotificationRequest), data
		)


	async def notifications(self,
					min_id: int | None = None,
					max_id: int | None = None,
					since_id: int | None = None,
					limit: int = 40,
					acct: Account | int | str | None = None,
					types: Sequence[NotificationType | str] | None = None,
					exclude: Sequence[NotificationType | str] | None = None) -> list[Notification]:

		data = parse_data(
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			limit = limit,
			account_id = object_to_id(acct),
			types = [NotificationType.parse(item) for item in types] if types else None,
			exclude_types = [NotificationType.parse(item) for item in exclude] if exclude else None
		)

		return await self.send("GET", "/api/v1/notifications", ListDeserializer(Notification), data)


	async def remove_all_notifications(self) -> None:
		await self.send("POST", "/api/v1/notifications/clear", None)


	async def remove_notification(self, notif: Notification | int | str) -> None:
		notif = object_to_id(notif)
		await self.send("POST", f"/api/v1/notifications/{notif}/dismiss", None)


	async def update_notification_policy(self,
										not_following: bool | None = None,
										not_followers: bool | None = None,
										new_accounts: bool | None = None,
										private_mentions: bool | None = None) -> NotificationPolicy:

		data = parse_data(
			filter_not_following = not_following,
			filter_not_followers = not_followers,
			filter_new_accounts = new_accounts,
			filter_private_mentions = private_mentions
		)

		return await self.send("PUT", "/api/v1/notifications/policy", NotificationPolicy, data)



from collections.abc import AsyncIterable, Sequence
from typing import Literal

from .base import ClientBase, object_to_id, parse_data

from ..enums import AccountStatus, AdminAccountAction, BlockSeverity
from ..objects import (
	AdminAccount,
	AdminDomainAllow,
	AdminDomainBlock,
	AdminEmailBlock,
	AdminReport,
	ListDeserializer,
	ObjectBase,
	Role
)


ORIGIN_VALUES = Literal["local", "remote"]


# remove when implemented
class AdminWarningPreset(ObjectBase):
	pass


class AdminBase(ClientBase):
	async def admin_account(self, id: str) -> AdminAccount:
		return await self.send("GET", f"/api/v1/admin/accounts/{id}", AdminAccount)


	async def admin_account_action(self,
								id: AdminAccount | str,
								type: AdminAccountAction | str,
								report: AdminReport | str | None = None,
								warning_preset: AdminWarningPreset | str | None = None,
								text: str | None = None,
								send_email_notification: bool = True) -> None:

		id = object_to_id(id)
		data = parse_data(
			type = AdminAccountAction.parse(type),
			report_id = object_to_id(report),
			warning_preset_id = object_to_id(warning_preset),
			text = text,
			send_email_notification = send_email_notification
		)

		await self.send("POST", f"/api/v1/admin/accounts/{id}/action", None, data)


	async def admin_account_action_revert(self,
										id: AdminAccount | str,
										type: AdminAccountAction | str) -> AdminAccount:

		id = object_to_id(id)

		match AdminAccountAction.parse(type):
			case AdminAccountAction.DISABLE:
				action = "enable"

			case AdminAccountAction.SENSITIVE:
				action = "unsensitive"

			case AdminAccountAction.SUSPEND:
				action = "unsuspend"

			case AdminAccountAction.SILENCE:
				action = "unsilence"

			case _:
				raise ValueError("Invalid account action")

		return await self.send("POST", f"/api/v1/admin/accounts/{id}/{action}", AdminAccount)


	async def admin_account_delete(self, id: AdminAccount | str) -> AdminAccount:
		id = object_to_id(id)
		return await self.send("DELETE", f"/api/v1/admin/accounts/{id}", AdminAccount)


	async def admin_account_response(self, id: AdminAccount | str, approve: bool) -> AdminAccount:
		id = object_to_id(id)
		response = "approve" if approve else "reject"
		return await self.send("POST", f"/api/v1/admin/accounts/{id}/{response}", AdminAccount)


	async def admin_accounts(self,
							origin: ORIGIN_VALUES | None = None,
							status: AccountStatus | str | None = None,
							permissions: str | None = None,
							roles: Sequence[Role | str] | None = None,
							invited_by: AdminAccount | str | None = None,
							username: str | None = None,
							display_name: str | None = None,
							domain: str | None = None,
							email: str | None = None,
							ip: str | None = None,
							max_id: str | None = None,
							min_id: str | None = None,
							since_id: str | None = None,
							limit: int = 100) -> list[AdminAccount]:

		data = parse_data(
			origin = origin,
			status = AccountStatus.parse(status) if status else None,
			permissions = permissions,
			role_ids = [object_to_id(v) for v in roles] if roles is not None else None,
			invited_by = object_to_id(invited_by),
			username = username,
			display_name = display_name,
			by_domain = domain,
			email = email,
			ip = ip,
			max_id = max_id,
			min_id = min_id,
			since_id = since_id,
			limit = limit
		)

		return await self.send(
			"GET", "/api/v2/admin/accounts", ListDeserializer(AdminAccount), data
		)


	# async def admin_dimensions() -> AdminDimension


	async def admin_block_domain(self,
						domain: str,
						severity: BlockSeverity | str = BlockSeverity.SILENCE,
						reject_media: bool = False,
						reject_reports: bool = False,
						private_comment: str | None = None,
						public_comment: str | None = None,
						obfuscate: bool = False) -> AdminDomainBlock:

		data = parse_data(
			domain = domain,
			severity = BlockSeverity.parse(severity),
			reject_media = reject_media,
			reject_reports = reject_reports,
			private_comment = private_comment,
			public_comment = public_comment,
			obfuscate = obfuscate
		)

		return await self.send("POST", "/api/v1/admin/domain_blocks", AdminDomainBlock, data)


	async def admin_allow_domain(self, domain: str) -> AdminDomainAllow:
		params = {"domain": domain}
		return await self.send("POST", "/api/v1/admin/domain_allows", AdminDomainAllow, params)


	async def admin_domain_allow(self, allow_id: int | str) -> AdminDomainAllow:
		return await self.send("GET", f"/api/v1/admin/domain_allows/{allow_id}", AdminDomainAllow)


	async def admin_domain_allows(self,
						min_id: int | None = None,
						max_id: int | None = None,
						since_id: int | None = None,
						limit: int = 100) -> list[AdminDomainAllow]:

		assert 0 <= limit <= 200, "Limit range must be 0 - 200"

		query = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send(
			"GET", "/api/v1/admin/domain_allows", ListDeserializer(AdminDomainAllow), query = query
		)


	async def admin_domain_block(self, block_id: int | str) -> AdminDomainBlock:
		return await self.send("GET", f"/api/v1/admin/domain_blocks/{block_id}", AdminDomainBlock)


	async def admin_domain_unblock(self, block_id: int | str) -> AdminDomainBlock:
		return await self.send(
			"DELETE", f"/api/v1/admin/domain_blocks/{block_id}", AdminDomainBlock
		)


	async def admin_domain_blocks(self,
						min_id: int | None = None,
						max_id: int | None = None,
						since_id: int | None = None,
						limit: int = 100) -> list[AdminDomainBlock]:

		assert 0 <= limit <= 200, "Limit range must be 0 - 200"

		query = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send(
			"GET", "/api/v1/admin/domain_blocks", ListDeserializer(AdminDomainBlock), query = query
		)


	async def admin_email_block(self, id: str) -> AdminEmailBlock:
		return await self.send("GET", f"/api/v1/admin/canonical_email_blocks/{id}", AdminEmailBlock)


	async def admin_email_block_add(self,
									email: str | None = None,
									hashdata: str | None = None) -> AdminEmailBlock:

		if not email and not hashdata:
			raise ValueError("Must provide email or hashdata")

		data = parse_data(
			email = email,
			canonical_email_hash = hashdata
		)

		return await self.send(
			"POST", "/api/v1/admin/canonical_email_blocks", AdminEmailBlock, data
		)


	async def admin_email_block_delete(self, id: AdminEmailBlock | str) -> None:
		id = object_to_id(id)
		return await self.send("DELETE", f"/api/v1/admin/canonical_email_blocks/{id}", None)


	async def admin_email_block_test(self, email: str) -> list[AdminEmailBlock]:
		return await self.send(
			"POST", "/api/v1/admin/canonical_email_blocks/test",
			ListDeserializer(AdminEmailBlock),
			{"email": email}
		)


	async def admin_email_blocks(self,
								max_id: str | None = None,
								min_id: str | None = None,
								limit: str | None = None,
								since_id: str | None = None) -> list[AdminEmailBlock]:

		return await self.send(
			"GET", "/api/v1/admin/canonical_email_blocks", ListDeserializer(AdminEmailBlock)
		)


	async def admin_update_domain(self,
						block: AdminDomainBlock | int | str,
						severity: BlockSeverity | str | None = None,
						reject_media: bool | None = None,
						reject_reports: bool | None = None,
						private_comment: str | None = None,
						public_comment: str | None = None,
						obfuscate: bool | None = None) -> AdminDomainBlock:

		block_id = object_to_id(block)
		data = parse_data(
			severity = BlockSeverity.parse(severity) if severity else None,
			reject_media = reject_media,
			reject_reports = reject_reports,
			private_comment = private_comment,
			public_comment = public_comment,
			obfuscate = obfuscate
		)

		return await self.send(
			"PUT", f"/api/v1/admin/domain_blocks/{block_id}", AdminDomainBlock, data
		)


	async def admin_unallow_domain(self, allow: AdminDomainAllow | int | str) -> AdminDomainAllow:
		return await self.send(
			"DELETE", f"/api/v1/admin/domain_allows/{object_to_id(allow)}", AdminDomainAllow
		)


	async def admin_unblock_domain(self, block: AdminDomainBlock | int | str) -> AdminDomainBlock:
		return await self.send(
			"DELETE", f"/api/v1/admin/domain_blocks/{object_to_id(block)}", AdminDomainBlock
		)


	# todo: create function based on this to work on any paginated endpoint
	async def all_admin_domain_blocks(self, chunk_size: int = 100) -> AsyncIterable[AdminDomainBlock]:
		max_id: int = 0

		while True:
			new_items = await self.admin_domain_blocks(max_id = max_id or None, limit = chunk_size)

			if not len(new_items):
				break

			max_id = int(new_items[-1].id)

			for item in new_items:
				yield item

			if len(new_items) < chunk_size:
				break



from blib import JsonBase, Url
from collections.abc import Sequence
from typing import Any

from .base import ClientBase, object_to_id, object_to_ids, parse_data

from ..enums import AccountAction, ReportCategory, VisibilityLevel
from ..objects import (
	Account,
	List,
	ListDeserializer,
	Relationship,
	Report,
	Rule,
	Status,
	Suggestion
)


class AccountBase(ClientBase):
	async def account(self, acct_id: int | str) -> Account:
		return await self.send("GET", f"/api/v1/accounts/{acct_id}", Account)


	# added in 4.3.0
	async def accounts(self, *acct_ids: str) -> Account:
		data = {"id": acct_ids}
		return await self.send("GET", "/api/v1/accounts", Account, data)


	async def account_action(self,
							acct: Account | int | str,
							action: AccountAction | str,
							**kwargs: Any) -> Relationship:

		acct = object_to_id(acct)
		action = AccountAction.parse(action)
		return await self.send("POST", f"/api/v1/accounts/{acct}/{action}", Relationship, kwargs)


	async def account_lookup(self, handle: str) -> Account | None:
		query = {"acct": handle}
		return await self.send("GET", "/api/v1/accounts/lookup", Account, query = query)


	async def block_domain(self, domain: str) -> None:
		await self.send("POST", "/api/v1/domain_blocks", None, {"domain": domain})


	async def blocked_accounts(self,
						min_id: int | None = None,
						max_id: int | None = None,
						since_id: int | None = None,
						limit: int = 20) -> list[Account]:

		query = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send("GET", "/api/v1/blocks", ListDeserializer(Account), query = query)


	async def blocked_domains(self,
						min_id: int | None = None,
						max_id: int | None = None,
						since_id: int | None = None,
						limit: int = 20) -> list[str]:

		query = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send("GET", "/api/v1/domain_blocks", ListDeserializer(None), query = query)


	async def create_account(self,
							username: str,
							email: str,
							password: str,
							agree: bool,
							locale: str = "en",
							reason: str | None = None) -> None:

		data = parse_data(
			username = username,
			email = email,
			password = password,
			agreement = agree,
			locale = locale,
			reason = reason
		)

		await self.send("POST", "/api/v1/accounts", None, data)


	async def delete_avatar(self) -> Account:
		return await self.send("DELETE", "/api/v1/profile/header", Account)


	async def familiar_followers(self, *accts: Account | str | int) -> list[Relationship]:
		data = {
			"id": [str(acct.id if isinstance(acct, Account) else acct) for acct in accts],
		}

		return await self.send(
			"GET", "/api/v1/accounts/familiar_followers", ListDeserializer(Relationship), data
		)


	async def feature(self, acct: Account | int | str) -> Relationship:
		return await self.account_action(acct, AccountAction.FEATURE)


	async def featured_accounts(self,
						limit: int = 40,
						max_id: int | None = None,
						since_id: int | None = None) -> list[Account]:

		data = parse_data(
			limit = limit,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send("GET", "/api/v1/endorsements", ListDeserializer(Account), data)


	async def follow(self,
					acct: Account | int | str,
					reblogs: bool = True,
					notify: bool = False,
					languages: Sequence[str] | None = None) -> Relationship:

		data = parse_data(
			reblogs = reblogs,
			notify = notify,
			languages = list(languages) if languages is not None else None
		) or {}

		return await self.account_action(acct, AccountAction.FOLLOW, **data)


	async def follow_requests(self,
							limit: int = 40,
							min_id: int | None = None,
							max_id: int | None = None) -> list[Account]:

		data = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id
		)

		return await self.send("GET", "/api/v1/follow_requests", ListDeserializer(Account), data)


	async def followers(self,
									acct: Account | int | str,
									min_id: int | None = None,
									max_id: int | None = None,
									since_id: int | None = None,
									limit: int = 40) -> list[Account]: # max 80

		acct = object_to_id(acct)
		data = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send(
			"GET", f"/api/v1/accounts/{acct}/followers", ListDeserializer(Account), data
		)


	async def following(self,
					acct: Account | int | str,
					min_id: int | None = None,
					max_id: int | None = None,
					since_id: int | None = None,
					limit: int = 40) -> list[Account]: # max 80

		acct = object_to_id(acct)
		data = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send(
			"GET", f"/api/v1/accounts/{acct}/following", ListDeserializer(Account), data
		)


	async def lists_containing_account(self, acct: Account | int | str) -> list[List]:
		acct = object_to_id(acct)
		return await self.send("GET", f"/api/v1/accounts/{acct}/lists", ListDeserializer(List))


	async def me(self) -> Account:
		return await self.send("GET", "/api/v1/accounts/verify_credentials", Account)


	async def mute_account(self, acct: Account | int | str) -> Relationship:
		return await self.account_action(acct, AccountAction.MUTE)


	async def muted_accounts(self,
						min_id: int | None = None,
						max_id: int | None = None,
						since_id: int | None = None,
						limit: int = 20) -> list[Account]:

		query = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send("GET", "/api/v1/mutes", ListDeserializer(Account), query = query)


	async def preferences(self) -> JsonBase[Any]:
		return await self.send("GET", "/api/v1/preferences", JsonBase)


	async def relationships(self,
						*accts: Account | str | int,
						with_suspended: bool = False) -> list[Relationship]:

		data = {
			"id": [str(acct.id if isinstance(acct, Account) else acct) for acct in accts],
			"with_suspended": with_suspended
		}

		return await self.send(
			"GET", "/api/v1/accounts/relationships", ListDeserializer(Relationship), data
		)


	async def remove_follower(self, acct: Account | int | str) -> Relationship:
		acct = object_to_id(acct)
		return await self.send(
			"POST", f"/api/v1/accounts/{acct}/remove_from_followers", Relationship
		)


	async def remove_suggested_account(self, acct: Account | int | str) -> None:
		acct = object_to_id(acct)
		await self.send("DELETE", f"/api/v2/suggestions/{acct}", None)


	async def report_user(self,
					acct: Account | int | str,
					statuses: list[Status | int | str] | None = None,
					rules: list[Rule | int | str] | None = None,
					comment: str | None = None,
					forward: bool = False,
					category: ReportCategory = ReportCategory.OTHER) -> Report:

		data = parse_data(
			account_id = object_to_id(acct),
			status_ids = object_to_ids(statuses or []) or None,
			comment = comment,
			forward = forward,
			category = category if not rules else ReportCategory.VIOLATION,
			rule_ids = object_to_ids(rules or []) or None
		)

		return await self.send("POST", "/api/v1/reports", Report, data)


	async def response_to_request(self, acct: Account | int | str, accept: bool) -> Relationship:
		if isinstance(acct, Account):
			acct = acct.id

		action = "authorize" if accept else "reject"
		return await self.send("POST", f"/api/v1/follow_requests/{acct}/{action}", Relationship)


	# resolve parameter does nothing
	async def search_accounts(self,
					query: Url | str,
					limit: int = 40,
					offset: int | None = None,
					following: bool = False) -> list[Account]:

		data = parse_data(
			q = query,
			limit = str(limit),
			offset = offset,
			following = str(following).lower(),
			resolve = str(isinstance(query, Url) or query.startswith(("https://", "http://"))).lower()
		)

		return await self.send(
			"GET", "/api/v1/accounts/search", ListDeserializer(Account), query = data
		)


	async def set_account_note(self, acct: Account | int | str, note: str | None) -> Relationship:
		acct = object_to_id(acct)
		params = {"comment": note or ""}
		return await self.send("POST", f"/api/v1/accounts/{acct}/note", Relationship, params)


	async def suggested_accounts(self, limit: int = 40) -> list[Suggestion]:
		data = {"limit": limit}
		return await self.send("GET", "/api/v2/suggestions", ListDeserializer(Suggestion), data)


	async def unblock_domain(self, domain: str) -> None:
		await self.send("DELETE", "/api/v1/domain_blocks", None, {"domain": domain})


	async def unfeature(self, acct: Account | int | str) -> Relationship:
		return await self.account_action(acct, AccountAction.UNFEATURE)


	async def unfollow(self, acct: Account | int | str) -> Relationship:
		return await self.account_action(acct, AccountAction.UNFOLLOW)


	async def unmute_account(self, acct: Account | int | str) -> Relationship:
		return await self.account_action(acct, AccountAction.UNMUTE)


	async def update_account(self,
							display_name: str | None = None,
							bio: str | None = None,
							avatar: Any = None,
							header: Any = None,
							locked: bool | None = None,
							bot: bool | None = None,
							discoverable: bool | None = None,
							hide_relationships: bool | None = None,
							indexable: bool | None = None,
							default_post_level: VisibilityLevel | str | None = None,
							default_sensitive: bool | None = None,
							default_language: str | None = None,
							fields: dict[str, str] | None = None) -> Account:

		if default_post_level is not None:
			if isinstance(default_post_level, str):
				default_post_level = VisibilityLevel.parse(default_post_level)

			if default_post_level == VisibilityLevel.DIRECT:
				raise ValueError("Default post level cannot be 'direct'")

		data = parse_data(
			display_name = display_name,
			note = bio,
			locked = locked,
			bot = bot,
			hide_collections = hide_relationships,
			indexable = indexable,
			source = parse_data(
				privacy = default_post_level,
				sensitive = default_sensitive
			)
		)

		if fields is not None:
			if data is None:
				data = {}

			data["fields_attributes"] = {}

			for idx, item in enumerate(fields.items()):
				data[str(idx)] = {
					"name": item[0],
					"value": item[1]
				}

		return await self.send("PATCH", "/api/v1/accounts/update_credentials", Account, data)



from .base import ClientBase, object_to_id, parse_data

from ..objects import Account, FeaturedTag, ListDeserializer, Tag


class TagBase(ClientBase):
	async def featured_account_tags(self, acct: Account | int | str) -> list[FeaturedTag]:
		acct = object_to_id(acct)
		return await self.send(
			"GET",
			f"/api/v1/accounts/{acct}/featured_tags",
			ListDeserializer[FeaturedTag](FeaturedTag)
		)


	async def featured_tags(self) -> list[FeaturedTag]:
		return await self.send("GET", "/api/v1/featured_tags", ListDeserializer(FeaturedTag))


	async def feature_tag(self, name: str) -> FeaturedTag:
		if name.startswith("#"):
			name = name[1:]

		return await self.send("POST", "/api/v1/featured_tags", FeaturedTag, {"name": name})


	async def follow_tag(self, tag: Tag | int | str) -> Tag:
		tag = object_to_id(tag, "name")
		return await self.send("POST", f"/api/v1/tags/{tag}/follow", Tag)


	async def followed_tags(self,
							limit: int = 100,
							min_id: int | None = None,
							max_id: int | None = None,
							since_id: int | None = None) -> list[Tag]:

		data = parse_data(
			limit = limit,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id
		)

		return await self.send("GET", "/api/v1/followed_tags", ListDeserializer(Tag), data)


	async def suggested_tags(self) -> list[FeaturedTag]:
		return await self.send(
			"GET", "/api/v1/featured_tags/suggestions", ListDeserializer(FeaturedTag)
		)


	async def tag(self, id: str) -> Tag:
		return await self.send("GET", f"/api/v1/tags/{id}", Tag)


	async def unfeature_tag(self, tag: FeaturedTag | int | str) -> None:
		tag = object_to_id(tag)
		await self.send("DELETE", f"/api/v1/feature_tags/{tag}", None)


	async def unfollow_tag(self, tag: Tag | int | str) -> Tag:
		tag = object_to_id(tag, "name")
		return await self.send("POST", f"/api/v1/tags/{tag}/unfollow", Tag)



from base64 import b64encode
from blib import File, Url
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Literal

from .base import ClientBase, object_to_id, parse_data

from ..enums import SearchType, WebPushPolicy
from ..objects import (
	Account,
	Announcement,
	Conversation,
	Emoji,
	LinkTrend,
	ListDeserializer,
	Markers,
	Media,
	OEmbed,
	Poll,
	Search,
	Status,
	Tag,
	WebPushSubscription
)


@dataclass(slots = True)
class SearchQuery:
	text: str | None = None
	has_type: Literal["media", "poll", "embed"] | None = None
	is_type: Literal["reply", "sensitive"] | None = None
	in_type: Literal["library", "public"] | None = None
	language: str | None = None
	from_acct: Account | int | str | None = None
	before: datetime | None = None
	during: datetime | None = None
	after: datetime | None = None


	def to_str(self) -> str:
		items: list[str] = []

		data = parse_data(
			has_type = self.has_type,
			is_type = self.is_type,
			in_type = self.in_type,
			language = self.language,
			from_acct = self.from_acct,
			before = self.before,
			during = self.during,
			after = self.after,
			text = self.text
		)

		if not data:
			raise ValueError("Must specify at least one search option")

		for key, value in data.items():
			key = key.split("_", 1)[0]

			if isinstance(value, datetime):
				value = value.strftime("%Y-%m-%d")

			elif isinstance(value, Account):
				value = value.acct

			items.append(value if key == "text" else f"{key}:{value}")

		return " ".join(items)


class MiscBase(ClientBase):
	async def announcement_reaction(self,
									item: Announcement | int | str,
									emoji: Emoji | str) -> None:

		data = parse_data(
			item = object_to_id(item),
			emoji = object_to_id(emoji, "name")
		)

		await self.send("PUT", f"/api/v1/announcements/{item}/reactions/{emoji}", None, data)


	async def announcements(self) -> list[Announcement]:
		return await self.send("GET", "/api/v1/announcements", ListDeserializer(Announcement))


	async def conversations(self,
							min_id: int | None = None,
							max_id: int | None = None,
							since_id: int | None = None,
							limit: int = 20) -> list[Conversation]:

		data = parse_data(
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			limit = limit
		)

		return await self.send(
			"GET", "/api/v1/conversations", ListDeserializer(Conversation), data
		)


	async def directory(self,
						limit: int = 40,
						offset: int | None = 0,
						order: Literal["active", "new"] = "active",
						local: bool = False) -> list[Account]:

		data = parse_data(
			limit = limit,
			offset = offset,
			order = order,
			local = local
		)

		return await self.send("GET", "/api/v1/directory", ListDeserializer(Account), data)


	async def dismiss_announcement(self, item: Announcement | int | str) -> None:
		item = object_to_id(item)
		return await self.send("POST", f"/api/v1/announcements/{item}/dissmiss", None)


	async def emojis(self) -> list[Emoji]:
		return await self.send(
			"GET", "/api/v1/custom_emojis", ListDeserializer(Emoji), token = False
		)


	async def mark_conversation_read(self, id: Conversation | int | str) -> Conversation:
		id = object_to_id(id)
		return await self.send("POST", f"/api/v1/conversations/{id}/read", Conversation)


	async def markers(self, home: bool = True, notifications: bool = True) -> Markers:
		if not home and not notifications:
			raise ValueError("Must specify at least one timeline")

		data = []

		if home:
			data.append("home")

		if notifications:
			data.append("notifications")

		return await self.send("GET", "/api/v1/markers", Markers, data)


	async def media(self, id: str | int) -> Media:
		return await self.send("GET", f"/api/v1/media/{id}", Media)


	async def oembed(self,
					url: Status | str,
					width: int | None = None,
					height: int | None = None) -> OEmbed:

		data = parse_data(
			url = object_to_id(url, "url"),
			width = width,
			height = height
		)

		return await self.send("GET", "/api/oembed", OEmbed, data, token = False)


	async def poll(self, id: str | int) -> Poll:
		return await self.send("GET", f"/api/v1/polls/{id}", Poll)


	async def push_subscribe(self,
							endpoint: str,
							pubkey: str,
							auth_secret: str,
							mention: bool = False,
							status: bool = False,
							boost: bool = False,
							follow: bool = False,
							follow_request: bool = False,
							favorite: bool = False,
							poll: bool = False,
							update: bool = False,
							sign_up: bool = False,
							report: bool = False,
							policy: WebPushPolicy | str = WebPushPolicy.ALL) -> WebPushSubscription:

		data = {
			"policy": policy,
			"subscription": {
				"endpoint": endpoint,
				"keys": {
					"p256dh": b64encode(pubkey.encode("utf8")).decode("utf-8"),
					"auth": auth_secret
				}
			},
			"data": {
				"mention": mention,
				"status": status,
				"reblog": boost,
				"follow": follow,
				"follow_request": follow_request,
				"favourite": favorite,
				"poll": poll,
				"update": update,
				"admin.sign_up": sign_up,
				"admin.report": report
			}
		}

		return await self.send("POST", "/api/v1/push/subscription", WebPushSubscription, data)


	async def push_subscription(self) -> WebPushSubscription:
		return await self.send("GET", "/api/v1/push/subscription", WebPushSubscription)


	async def remove_announcement_reaction(self,
									item: Announcement | int | str,
									emoji: Emoji | str) -> None:

		data = parse_data(
			item = object_to_id(item),
			emoji = object_to_id(emoji, "name")
		)

		await self.send("DELETE", f"/api/v1/announcements/{item}/reactions/{emoji}", None, data)


	async def remove_conversation(self, id: Conversation | int | str) -> None:
		id = object_to_id(id)
		await self.send("DELETE", f"/api/v1/conversations/{id}", None)


	async def remove_push_subscription(self) -> None:
		await self.send("DELETE", "/api/v1/push/subscription", None)


	async def search(self,
					query: SearchQuery | Url | str,
					type: SearchType | str | None = None,
					following: bool = False,
					acct: Account | int | str | None = None,
					exclude_unreviewed: bool = False,
					min_id: int | str | None = None,
					max_id: int | str | None = None,
					limit: int = 20,
					offset: int | None = None) -> Search | list[Account] | list[Tag] | list[Status]:

		type = SearchType.parse(type) if type is not None else type
		data = parse_data(
			q = query.to_str() if isinstance(query, SearchQuery) else query,
			type = type,
			following = following,
			account_id = object_to_id(acct),
			exclude_unreviewed = exclude_unreviewed,
			min_id = min_id,
			max_id = max_id,
			limit = limit,
			offset = offset,
			resolve = isinstance(query, str) and query.startswith(("https://", "http://"))
		)

		resp = await self.send("GET", "/api/v2/search", Search, data)

		if type == SearchType.ACCOUNTS:
			return resp.accounts

		if type == SearchType.HASHTAGS:
			return resp.hashtags

		if type == SearchType.STATUSES:
			return resp.statuses

		return resp


	async def set_marker(self,
						home: Status | int | str | None = None,
						notifications: Status | int | str | None = None) -> Markers:

		if home is None and notifications is None:
			raise ValueError("At least one timeline must be specified")

		data = {}

		if home is not None:
			data["home"] = {"last_read_id": object_to_id(home)}

		if notifications is not None:
			data["notifications"] = {"last_read_id": object_to_id(notifications)}

		return await self.send("POST", "/api/v1/markers", Markers, data)


	async def trending_links(self, limit: int = 10, offset: int = 0) -> list[LinkTrend]:
		data = {
			"limit": limit,
			"offset": offset
		}

		return await self.send("GET", "/api/v1/trends/links", ListDeserializer(LinkTrend), data)


	async def trending_statuses(self, limit: int = 10, offset: int = 0) -> list[Status]:
		data = {
			"limit": limit,
			"offset": offset
		}

		return await self.send("GET", "/api/v1/trends/statuses", ListDeserializer(Status), data)


	async def trending_tags(self, limit: int = 10, offset: int = 0) -> list[Tag]:
		data = {
			"limit": limit,
			"offset": offset
		}

		return await self.send("GET", "/api/v1/trends/tags", ListDeserializer(Tag), data)


	async def update_media(self,
						media: Media | int | str,
						thumbnail: File | Path | str | None = None,
						description: str | None = None,
						focus: tuple[float, float] | None = None) -> Media:

		media = object_to_id(media)

		if isinstance(thumbnail, (Path, str)):
			thumbnail = Path(thumbnail)

		data = parse_data(
			thumbnail = thumbnail.resolve() if thumbnail is not None else None,
			description = description,
			focus = ",".join(str(item) for item in focus) if focus is not None else None
		)

		return await self.send("PUT", f"/api/v2/media/{media}", Media, form = data)


	async def update_push_subscription(self,
							mention: bool = False,
							status: bool = False,
							boost: bool = False,
							follow: bool = False,
							follow_request: bool = False,
							favorite: bool = False,
							poll: bool = False,
							update: bool = False,
							sign_up: bool = False,
							report: bool = False,
							policy: WebPushPolicy | str = WebPushPolicy.ALL) -> WebPushSubscription:

		data = {
			"policy": policy,
			"data": {
				"mention": mention,
				"status": status,
				"reblog": boost,
				"follow": follow,
				"follow_request": follow_request,
				"favourite": favorite,
				"poll": poll,
				"update": update,
				"admin.sign_up": sign_up,
				"admin.report": report
			}
		}

		return await self.send("PUT", "/api/v1/push/subscription", WebPushSubscription, data)


	async def upload_media(self,
						path: File | Path | str,
						thumbnail: File | Path | str | None = None,
						description: str | None = None,
						focus: tuple[float, float] | None = None) -> Media:

		data = parse_data(
			file = File(str(path)).resolve(),
			thumbnail = File(str(thumbnail)).resolve() if thumbnail is not None else None,
			description = description,
			focus = ",".join(str(item) for item in focus) if focus is not None else None
		)

		return await self.send("POST", "/api/v2/media", Media, form = data)


	# todo: make sure this is how the endpoint actually works
	async def vote(self, poll: Poll | int | str, *choices: int) -> Poll:
		poll = object_to_id(poll)
		return await self.send("POST", f"/api/v1/polls/{poll}/votes", Poll, list(choices))



import webbrowser

from blib import Query, Url

from .base import ClientBase, parse_data

from ..objects import AccessToken, Application


class AppBase(ClientBase):
	def authorization_url(self,
							client_id: str | None = None,
							redirect_uri: str | None = None,
							scopes: list[str] | None = None,
							force_login: bool = False,
							lang: str = "EN") -> Url:

		if client_id is None:
			client_id = self.client_id

		assert client_id is not None, "'client_id' must be set"

		query = Query({
			"response_type": "code",
			"client_id": client_id,
			"redirect_uri": redirect_uri or self.redirect_uri,
			"scope": " ".join(scopes or self.scopes),
			"force_login": "true" if force_login else "false",
			"lang": lang
		})

		return Url(
			domain = self.host,
			path = "/oauth/authorize",
			proto = "https" if self.https else "http",
			query = query
		)


	async def create_app(self,
						name: str | None = None,
						redirect_uri: str | None = None,
						scopes: list[str] | None = None,
						website: str | None = None,
						set_app_details: bool = False) -> Application:

		data = {
			"client_name": name or self.name,
			"redirect_uris": redirect_uri or self.redirect_uri,
			"scopes": " ".join(scopes if scopes else self.scopes),
			"website": website or self.website
		}

		assert data["client_name"] is not None, "'client_name' must be set"

		app = await self.send("POST", "/api/v1/apps", Application, data, token = False)
		app.scope = data["scopes"] # type: ignore[assignment]

		if set_app_details:
			self.set_details_from_app(app)

		return app


	async def obtain_token(self,
						client_id: str | None = None,
						client_secret: str | None = None,
						redirect_uri: str | None = None,
						scopes: list[str] | None = None,
						code: str | None = None,
						set_token: bool = False) -> AccessToken:

		data = parse_data(
			grant_type = "authorization_code" if code is not None else "client_credentials",
			client_id = client_id or self.client_id,
			client_secret = client_secret or self.client_secret,
			redirect_uri = redirect_uri or self.redirect_uri,
			scope = " ".join(scopes or self.scopes),
			code = code
		)

		assert data is not None, "Must set 'client_id' and 'client_secret'"
		assert None not in [data.get("client_id"), data.get("client_secret")], \
			"Must set 'client_id' and 'client_secret'"

		token = await self.send("POST", "/oauth/token", AccessToken, data, token = False)

		if set_token:
			self.token = token.access_token

		return token


	def open_authorization_url(self,
							client_id: str | None = None,
							redirect_uri: str | None = None,
							scopes: list[str] | None = None,
							force_login: bool = False,
							lang: str = "EN") -> None:

		url = self.authorization_url(client_id, redirect_uri, scopes, force_login, lang)
		webbrowser.open(url)


	async def resend_confirmation(self, email: str) -> None:
		await self.send("POST", "/api/v1/emails/confirmations", None, {"email": email})


	async def revoke_token(self, id: str, secret: str, token: str) -> None:
		data = parse_data(
			client_id = id,
			client_secret = secret,
			token = token
		)

		await self.send("POST", "/oauth/revoke", None, data)


	async def verify_token(self) -> Application:
		return await self.send("GET", "/api/v1/apps/verify_credentials", Application)



__software__ = "Asynctodon"
__version_info__ = (0, 1, 5)
__version__ = ".".join(str(v) for v in __version_info__)


from .api.base import ClientBase, StreamEvent, StreamParser
from .api.misc import SearchQuery
from .base import ApiDate
from .client import Client

from .enums import (
	AccountAction,
	AccountStatus,
	AdminAccountAction,
	AdminReportCategory,
	BlockSeverity,
	FilterAction,
	GrantType,
	ListReplyPolicy,
	MediaType,
	MuteContext,
	NotificationType,
	PermissionFlag,
	PreviewCardType,
	ReportCategory,
	SearchType,
	SeveredRelationshipType,
	StatusAction,
	StreamEventType,
	SuggestionType,
	VisibilityLevel,
	WebPushPolicy
)

from .errors import (
	ClientError,
	InvalidFieldError,
	ServerError
)

from .objects import (
	Field,
	ListDeserializer,
	ListSerializer,
	ObjectBase,

	# sub-objects
	AccountField,
	AdminEmailBlockHistory,
	AdminIP,
	AnnouncementAccount,
	AnnouncementStatus,
	CredentialSource,
	Emoji,
	Filter,
	FilterKeyword,
	FilterResult,
	FilterStatus,
	InstanceConfig,
	InstanceConfigAccounts,
	InstanceConfigMedia,
	InstanceConfigPoll,
	InstanceConfigStatus,
	InstanceConfigTranslation,
	InstanceConfigUrls,
	InstanceConfigVapid,
	InstanceContact,
	InstanceRegistration,
	InstanceThumbnail,
	InstanceUsage,
	InstanceUsageUsers,
	Marker,
	Media,
	Mention,
	NotificationGroup,
	NotificationPolicySummary,
	PartialAccount,
	Poll,
	PollOption,
	PreviewCard,
	Reaction,
	RelationshipSeveranceEvent,
	Role,
	StatusApplication,
	StatusTag,
	Tag,
	TagHistory,
	TranslationMedia,
	TranslationPoll,
	WebPushSubscriptionAlerts,

	# main objects
	AccessToken,
	Account,
	Activity,
	AdminAccount,
	AdminDomainAllow,
	AdminDomainBlock,
	AdminEmailBlock,
	AdminReport,
	Announcement,
	Application,
	Conversation,
	DomainBlock,
	ExtendedDescription,
	FeaturedTag,
	GroupedNotificationsResults,
	Instance,
	LinkTrend,
	LinkTrendHistory,
	List,
	Markers,
	Notification,
	NotificationPolicy,
	NotificationRequest,
	OEmbed,
	Relationship,
	Report,
	Rule,
	Search,
	ScheduledStatus,
	ScheduledStatusParams,
	ScheduledStatusParamsPoll,
	Status,
	StatusContext,
	StatusEdit,
	StatusSource,
	Suggestion,
	Translation,
	WebPushSubscription
)



from .base import ClientBase, object_to_id, parse_data

from ..enums import ListReplyPolicy
from ..objects import Account, List, ListDeserializer


class ListBase(ClientBase):
	async def create_list(self,
						title: str,
						replies_policy: ListReplyPolicy = ListReplyPolicy.LIST,
						exclusive: bool = False) -> List:

		data = {
			"title": title,
			"replies_policy": replies_policy,
			"exclusive": exclusive
		}

		return await self.send("POST", "/api/v1/lists", List, data = data)


	async def delete_list(self, id: List | int | str) -> None:
		id = object_to_id(id)
		return await self.send("DELETE", f"/api/v1/lists/{id}", None)


	async def get_list(self, id: int | str) -> List:
		return await self.send("GET", f"/api/v1/lists/{id}", List)


	async def list_add_accounts(self, id: List | int | str, *accts: Account | int | str) -> None:
		if not accts:
			raise ValueError("Must specify at least one account")

		id = object_to_id(id)
		data = [object_to_id(acct) for acct in accts]
		return await self.send("POST", f"/api/v1/lists/{id}/accounts", None, data)


	async def list_remove_accounts(self, id: List | int | str, *accts: Account | int | str) -> None:
		if not accts:
			raise ValueError("Must specify at least one account")

		id = object_to_id(id)
		data = [object_to_id(acct) for acct in accts]
		return await self.send("DELETE", f"/api/v1/lists/{id}/accounts", None, data)


	async def list_accounts(self,
							id: List | int | str,
							min_id: int | None = None,
							max_id: int | None = None,
							since_id: int | None = None,
							limit: int = 40) -> list[Account]:

		id = object_to_id(id)
		data = parse_data(
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			limit = limit
		)

		return await self.send(
			"PUT", f"/api/v1/lists/{id}/accounts", ListDeserializer(Account), data
		)


	async def lists(self) -> list[List]:
		return await self.send("GET", "/api/v1/lists", ListDeserializer(List))


	async def update_list(self,
						id: List | int | str,
						title: str | None,
						replies_policy: ListReplyPolicy | None = None,
						exclusive: bool | None = None) -> List:

		id = object_to_id(id)
		data = parse_data(
			title = title,
			replies_policy = replies_policy,
			exclusive = exclusive
		)

		if not data:
			raise ValueError("Must specify at least one parameter to change")

		return await self.send("PUT", f"/api/v1/lists/{id}", List, data)
